# noinspection PyUnresolvedReferences
from pyscript import document, window
# noinspection PyUnresolvedReferences
from pyscript.ffi import create_proxy
from functools import wraps

from visualgo.logic import ControllerInterface, Controller, PyDebugger, SymbolDescription

from pyscript_ui_callbacks import PyscriptUICallbacks
import pyscript_to_worker

# noinspection PyUnresolvedReferences
from pyscript.js_modules import ui_public_functions as ui

from program_state import ExecutionState


def initialize_if_needed_async(func):
    """
    Decorator to check if the Controller was initialized before calling a method.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if ExecutionState.is_not_initialized():
            ui.disableCodeInput()
            ExecutionState.set_to_paused()
            self.controller.initialize()
        return await func(self, *args, **kwargs)

    return wrapper


def initialize_if_needed(func):
    """
    Decorator to check if the Controller was initialized before calling a method.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if ExecutionState.is_not_initialized():
            ui.disableCodeInput()
            ExecutionState.set_to_paused()
            self.controller.initialize()
        return func(self, *args, **kwargs)

    return wrapper


class Visualisation():
    """
    This class is the main interface between the pyscript visualisation and the Controller.
    """

    def __init__(self):
        self.controller = Controller(PyDebugger, PyscriptUICallbacks())
        ExecutionState.set_to_not_initialized()

    # Execution control

    @initialize_if_needed_async
    async def step_forward_automatic(self, _):
        ui.disableInteractions()
        ExecutionState.set_to_running()
        # print("calling step_forward_automatic")
        await self.controller.step_forward_automatic()

    def pause(self, _):
        if ExecutionState.is_running():
            self.controller.pause()
            ui.enableButtons()
            ExecutionState.set_to_paused()

    @initialize_if_needed
    def forward_step(self, _):
        self.controller.forward_step()

    @initialize_if_needed
    def do_continue(self, _):
        self.controller.do_continue()

    def backward_step(self, _):
        # ui.enableButtons() # TODO uncomment when it is implemented in Debugger
        self.controller.backward_step()

    def stop(self, _):
        self.controller.stop()

    def set_step_time(self, time_ms):
        self.controller.set_step_time(time_ms)

    # Checkpoints and Breakpoints

    def new_checkpoint(self, line_number: int):
        self.controller.new_checkpoint(line_number)

    def del_checkpoint(self, line_number: int):
        self.controller.del_checkpoint(line_number)

    def new_breakpoint(self, line_number: int):
        self.controller.new_breakpoint(line_number, "")

    def del_breakpoint(self, line_number: int):
        self.controller.del_breakpoint(line_number)

    def new_tracker(self, name: str, line_number: int):
        self.controller.new_tracker(name, line_number)

    def del_tracker(self, name: str, line_number: int):
        self.controller.del_tracker(name, line_number)

    # For tracking for drawings

    def get_static_variables_and_user_fun(self, _):
        return self.controller.get_static_variables_and_user_fun()

    def hide_variable(self, _):
        self.controller.hide_variable(SymbolDescription("var_name", "function", 0))

    def show_variable(self, _):
        self.controller.show_variable(SymbolDescription("var_name", "function", 0))

    def hide_variables_in_user_fun(self, _):
        self.controller.hide_variables_in_user_fun(SymbolDescription("my_fun", "my_fun", 0))

    def show_variables_in_user_fun(self, _):
        self.controller.show_variables_in_user_fun(SymbolDescription("my_fun", "my_fun", 0))

    # Tracking for statistics

    def get_static_types_and_functions_calls(self, _):
        return self.controller.get_static_types_and_functions_calls()

    def new_tracked_function(self, _):
        self.controller.new_tracked_function("func_name")

    def del_tracked_function(self, _):
        self.controller.del_tracked_function("func_name")

    def new_tracked_type(self, _):
        self.controller.new_tracked_type("type_name")

    def del_tracked_type(self, _):
        self.controller.del_tracked_type("type_name")

    def get_csv(self, _):
        print(self.controller.get_csv())


if __name__ == "__main__":
    pyscript_to_worker.set_implementation()
    visualisation = Visualisation()

    # Execution control

    document.getElementById("startButton").addEventListener(
        "click", create_proxy(visualisation.step_forward_automatic)
    )

    document.getElementById("pauseButton").addEventListener(
        "click", create_proxy(visualisation.pause)
    )

    document.getElementById("stopButton").addEventListener(
        "click", create_proxy(visualisation.stop)
    )

    # document.getElementById("stepBackwardButton").addEventListener(
    #     "click", create_proxy(visualisation.backward_step)
    # )

    document.getElementById("stepForwardButton").addEventListener(
        "click", create_proxy(visualisation.forward_step)
    )

    document.getElementById("continueButton").addEventListener(
        "click", create_proxy(visualisation.do_continue)
    )

    # Checkpoints and Breakpoints
    window.set_step_time = create_proxy(visualisation.set_step_time)

    window.new_breakpoint = create_proxy(visualisation.new_breakpoint)
    window.del_breakpoint = create_proxy(visualisation.del_breakpoint)
    window.new_checkpoint = create_proxy(visualisation.new_checkpoint)
    window.del_checkpoint = create_proxy(visualisation.del_checkpoint)

    # Tracking for drawings

    document.getElementById("selectVariablesButton").addEventListener(
        "click", create_proxy(visualisation.get_static_variables_and_user_fun)
    )

    # Tracking for statistics

    window.new_tracker = create_proxy(visualisation.new_tracker)
    window.del_tracker = create_proxy(visualisation.del_tracker)

    document.getElementById("fetchStats").addEventListener(
        "click", create_proxy(visualisation.get_csv)
    )

    # disable loading screen.
    document.getElementById("loadingScreen").style.display = "none"

    # document.getElementById("exportStats").addEventListener(
    #     "click", create_proxy(visualisation.get_csv)
    # )
